@@ -23,13 +23,7 @@
 #ifndef _CHECK_H
 #define _CHECK_H
 
-off_t alloc_rootdir_entry(DOS_FS * fs, DIR_ENT * de, const char *pattern, int gen_name);
-
-/* Allocate a free slot in the root directory for a new file. If gen_name is
-   true, the file name is constructed after 'pattern', which must include a %d
-   type format for printf and expand to exactly 11 characters. The name
-   actually used is written into the 'de' structure, the rest of *de is cleared.
-   The offset returned is to where in the filesystem the entry belongs. */
+void check_dirty_bits(DOS_FS * fs);
 
 int scan_root(DOS_FS * fs);
 
@@ -37,4 +31,10 @@ int scan_root(DOS_FS * fs);
    for all the details. Returns a non-zero integer if the filesystem has to
    be checked again. */
 
+
+void check_label(DOS_FS * fs);
+
+/* Checks the volume label from the root directory entry that is valid and
+ * matches the label stored in boot sector. */
+
 #endif
