@@ -2,6 +2,8 @@
 
    Copyright (C) 1993 Werner Almesberger <werner.almesberger@lrc.di.epfl.ch>
    Copyright (C) 1998 Roman Hodek <Roman.Hodek@informatik.uni-erlangen.de>
+   Copyright (C) 2008-2014 Daniel Baumann <mail@daniel-baumann.ch>
+   Copyright (C) 2021 Pali Rohár <pali.rohar@gmail.com>
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -16,7 +18,7 @@
    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
 
-   On Debian systems, the complete text of the GNU General Public License
+   The complete text of the GNU General Public License
    can be found in /usr/share/common-licenses/GPL-3 file.
 */
 
@@ -29,8 +31,9 @@
 #include <unistd.h>
 
 #include "common.h"
-#include "dosfsck.h"
+#include "fsck.fat.h"
 #include "io.h"
+#include "boot.h"
 #include "check.h"
 #include "fat.h"
 
@@ -42,10 +45,15 @@
  * @param[in]	cluster     Cluster of interest
  * @param[in]	fs          Information from the FAT boot sectors (bits per FAT entry)
  */
-void get_fat(FAT_ENTRY * entry, void *fat, unsigned long cluster, DOS_FS * fs)
+void get_fat(FAT_ENTRY * entry, void *fat, uint32_t cluster, DOS_FS * fs)
 {
     unsigned char *ptr;
 
+    if (cluster > fs->data_clusters + 1) {
+	die("Internal error: cluster out of range in get_fat() (%lu > %lu).",
+		(unsigned long)cluster, (unsigned long)(fs->data_clusters + 1));
+    }
+
     switch (fs->fat_bits) {
     case 12:
 	ptr = &((unsigned char *)fat)[cluster * 3 / 2];
@@ -53,13 +61,13 @@ void get_fat(FAT_ENTRY * entry, void *fat, unsigned long cluster, DOS_FS * fs)
 				(ptr[0] | ptr[1] << 8));
 	break;
     case 16:
-	entry->value = CF_LE_W(((unsigned short *)fat)[cluster]);
+	entry->value = le16toh(((unsigned short *)fat)[cluster]);
 	break;
     case 32:
 	/* According to M$, the high 4 bits of a FAT32 entry are reserved and
 	 * are not part of the cluster number. So we cut them off. */
 	{
-	    unsigned long e = CF_LE_L(((unsigned int *)fat)[cluster]);
+	    uint32_t e = le32toh(((unsigned int *)fat)[cluster]);
 	    entry->value = e & 0xfffffff;
 	    entry->reserved = e >> 28;
 	}
@@ -69,6 +77,31 @@ void get_fat(FAT_ENTRY * entry, void *fat, unsigned long cluster, DOS_FS * fs)
     }
 }
 
+void release_fat(DOS_FS * fs)
+{
+    if (fs->fat)
+	free(fs->fat);
+    if (fs->cluster_owner)
+	free(fs->cluster_owner);
+    fs->fat = NULL;
+    fs->cluster_owner = NULL;
+}
+
+static void fix_first_cluster(DOS_FS * fs, void * first_cluster)
+{
+    struct boot_sector b;
+
+    fs_read(0, sizeof(b), &b);
+
+    printf("Fixing first cluster in FAT.\n");
+    if (fs->fat_bits == 12)
+        *(uint16_t *)first_cluster = htole16((le16toh(*(uint16_t *)first_cluster) & 0xf000) | FAT_EXTD(fs) | b.media);
+    else if (fs->fat_bits == 16)
+        *(uint16_t *)first_cluster = htole16(FAT_EXTD(fs) | b.media);
+    else
+        *(uint32_t *)first_cluster = htole32(FAT_EXTD(fs) | b.media);
+}
+
 /**
  * Build a bookkeeping structure from the partition's FAT table.
  * If the partition has multiple FATs and they don't agree, try to pick a winner,
@@ -76,64 +109,111 @@ void get_fat(FAT_ENTRY * entry, void *fat, unsigned long cluster, DOS_FS * fs)
  * One error that is fixed here is a cluster that links to something out of range.
  *
  * @param[inout]    fs      Information about the filesystem
+ * @param[in]       mode    0 - read-only, 1 - read-write (no repair), 2 - repair
  */
-void read_fat(DOS_FS * fs)
+void read_fat(DOS_FS * fs, int mode)
 {
-    int eff_size;
-    unsigned long i;
+    int eff_size, alloc_size;
+    uint32_t i;
     void *first, *second = NULL;
-    int first_ok, second_ok;
-    unsigned long total_num_clusters;
+    int first_ok, second_ok = 0;
+    FAT_ENTRY first_media, second_media;
+    uint32_t total_num_clusters;
+
+    if (fat_table > fs->nfats)
+        die("Requested FAT table %ld does not exist.", fat_table);
+    if (fat_table > 2)
+        die("Reading FAT table greather than 2 is implemented yet.");
 
     /* Clean up from previous pass */
-    free(fs->fat);
-    free(fs->cluster_owner);
-    fs->fat = NULL;
-    fs->cluster_owner = NULL;
+    release_fat(fs);
 
-    total_num_clusters = fs->clusters + 2UL;
+    total_num_clusters = fs->data_clusters + 2;
     eff_size = (total_num_clusters * fs->fat_bits + 7) / 8ULL;
-    first = alloc(eff_size);
+
+    if (fs->fat_bits != 12)
+	    alloc_size = eff_size;
+    else
+	    /* round up to an even number of FAT entries to avoid special
+	     * casing the last entry in get_fat() */
+	    alloc_size = (total_num_clusters * 12 + 23) / 24 * 3;
+
+    first = alloc(alloc_size);
     fs_read(fs->fat_start, eff_size, first);
+    get_fat(&first_media, first, 0, fs);
+    first_ok = (first_media.value & FAT_EXTD(fs)) == FAT_EXTD(fs);
     if (fs->nfats > 1) {
-	second = alloc(eff_size);
+	second = alloc(alloc_size);
 	fs_read(fs->fat_start + fs->fat_size, eff_size, second);
-    }
-    if (second && memcmp(first, second, eff_size) != 0) {
-	FAT_ENTRY first_media, second_media;
-	get_fat(&first_media, first, 0, fs);
 	get_fat(&second_media, second, 0, fs);
-	first_ok = (first_media.value & FAT_EXTD(fs)) == FAT_EXTD(fs);
 	second_ok = (second_media.value & FAT_EXTD(fs)) == FAT_EXTD(fs);
+    }
+    if (mode != 0 && fat_table == 0) {
+        if (!first_ok && second && !second_ok)
+            die("Both FATs appear to be corrupt. Giving up. Run fsck.fat with non-zero -F option.");
+        if (!first_ok && !second)
+            die("First FAT appears to be corrupt and second FAT does not exist. Giving up. Run fsck.fat with -F 1 option.");
+    }
+    if (mode == 0 && !first_ok && second && second_ok) {
+        /* In read-only mode if first FAT is corrupted and second is OK then use second FAT */
+        void *first_backup = first;
+        first = second;
+        second = first_backup;
+    }
+    if (mode != 0 && fat_table == 0 && second && memcmp(first, second, eff_size) != 0) {
+	if (mode != 2)
+	    die("FATs differ, please run fsck.fat");
 	if (first_ok && !second_ok) {
 	    printf("FATs differ - using first FAT.\n");
 	    fs_write(fs->fat_start + fs->fat_size, eff_size, first);
-	}
-	if (!first_ok && second_ok) {
+	} else if (!first_ok && second_ok) {
 	    printf("FATs differ - using second FAT.\n");
 	    fs_write(fs->fat_start, eff_size, second);
 	    memcpy(first, second, eff_size);
-	}
-	if (first_ok && second_ok) {
-	    if (interactive) {
-		printf("FATs differ but appear to be intact. Use which FAT ?\n"
-		       "1) Use first FAT\n2) Use second FAT\n");
-		if (get_key("12", "?") == '1') {
-		    fs_write(fs->fat_start + fs->fat_size, eff_size, first);
-		} else {
-		    fs_write(fs->fat_start, eff_size, second);
-		    memcpy(first, second, eff_size);
+	} else {
+	    if (first_ok && second_ok)
+		printf("FATs differ but appear to be intact.\n");
+	    else
+		printf("FATs differ and both appear to be corrupt.\n");
+	    if (get_choice(1, "  Using first FAT.",
+			   2,
+			   1, "Use first FAT",
+			   2, "Use second FAT") == 1) {
+		if (!first_ok) {
+		    fix_first_cluster(fs, first);
+		    fs_write(fs->fat_start, (fs->fat_bits + 7) / 8, first);
 		}
-	    } else {
-		printf("FATs differ but appear to be intact. Using first "
-		       "FAT.\n");
 		fs_write(fs->fat_start + fs->fat_size, eff_size, first);
+	    } else {
+		if (!second_ok) {
+		    fix_first_cluster(fs, second);
+		    fs_write(fs->fat_start + fs->fat_size, (fs->fat_bits + 7) / 8, second);
+		}
+		fs_write(fs->fat_start, eff_size, second);
+		memcpy(first, second, eff_size);
 	    }
 	}
-	if (!first_ok && !second_ok) {
-	    printf("Both FATs appear to be corrupt. Giving up.\n");
-	    exit(1);
-	}
+    }
+    if (mode != 0 && fat_table != 0) {
+        if (fat_table == 1) {
+            printf("Using first FAT.\n");
+            if (!first_ok) {
+                fix_first_cluster(fs, first);
+                fs_write(fs->fat_start, (fs->fat_bits + 7) / 8, first);
+            }
+            if (second && memcmp(first, second, eff_size) != 0)
+                fs_write(fs->fat_start + fs->fat_size, eff_size, first);
+        } else if (fat_table == 2) {
+            printf("Using second FAT.\n");
+            if (!second_ok) {
+                fix_first_cluster(fs, second);
+                fs_write(fs->fat_start + fs->fat_size, (fs->fat_bits + 7) / 8, second);
+            }
+            if (memcmp(first, second, eff_size) != 0) {
+                fs_write(fs->fat_start, eff_size, second);
+                memcpy(first, second, eff_size);
+            }
+        }
     }
     if (second) {
 	free(second);
@@ -143,18 +223,30 @@ void read_fat(DOS_FS * fs)
     fs->cluster_owner = alloc(total_num_clusters * sizeof(DOS_FILE *));
     memset(fs->cluster_owner, 0, (total_num_clusters * sizeof(DOS_FILE *)));
 
+    if (mode == 0)
+        return;
+
     /* Truncate any cluster chains that link to something out of range */
-    for (i = 2; i < fs->clusters + 2; i++) {
+    for (i = 2; i < fs->data_clusters + 2; i++) {
 	FAT_ENTRY curEntry;
 	get_fat(&curEntry, fs->fat, i, fs);
 	if (curEntry.value == 1) {
-	    printf("Cluster %ld out of range (1). Setting to EOF.\n", i - 2);
+	    if (mode != 2)
+		die("Cluster %ld out of range (1), please run fsck.fat",
+		    (long)(i - 2));
+	    printf("Cluster %ld out of range (1). Setting to EOF.\n",
+		   (long)(i - 2));
 	    set_fat(fs, i, -1);
 	}
-	if (curEntry.value >= fs->clusters + 2 &&
+	if (curEntry.value >= fs->data_clusters + 2 &&
 	    (curEntry.value < FAT_MIN_BAD(fs))) {
+	    if (mode != 2)
+		die("Cluster %ld out of range (%ld > %ld), please run fsck.fat",
+		    (long)(i - 2), (long)curEntry.value,
+		    (long)(fs->data_clusters + 2 - 1));
 	    printf("Cluster %ld out of range (%ld > %ld). Setting to EOF.\n",
-		   i - 2, curEntry.value, fs->clusters + 2 - 1);
+		   (long)(i - 2), (long)curEntry.value,
+		   (long)(fs->data_clusters + 2 - 1));
 	    set_fat(fs, i, -1);
 	}
     }
@@ -173,16 +265,26 @@ void read_fat(DOS_FS * fs)
  *				  -1 == end-of-chain
  *				  -2 == bad cluster
  */
-void set_fat(DOS_FS * fs, unsigned long cluster, unsigned long new)
+void set_fat(DOS_FS * fs, uint32_t cluster, int32_t new)
 {
     unsigned char *data = NULL;
     int size;
-    loff_t offs;
+    off_t offs;
 
-    if ((long)new == -1)
+    if (cluster > fs->data_clusters + 1) {
+	die("Internal error: cluster out of range in set_fat() (%lu > %lu).",
+		(unsigned long)cluster, (unsigned long)(fs->data_clusters + 1));
+    }
+
+    if (new == -1)
 	new = FAT_EOF(fs);
     else if ((long)new == -2)
 	new = FAT_BAD(fs);
+    else if (new > fs->data_clusters + 1) {
+	die("Internal error: new cluster out of range in set_fat() (%lu > %lu).",
+		(unsigned long)new, (unsigned long)(fs->data_clusters + 1));
+    }
+
     switch (fs->fat_bits) {
     case 12:
 	data = fs->fat + cluster * 3 / 2;
@@ -194,17 +296,19 @@ void set_fat(DOS_FS * fs, unsigned long cluster, unsigned long new)
 	    data[1] = new >> 4;
 	} else {
 	    FAT_ENTRY subseqEntry;
-	    get_fat(&subseqEntry, fs->fat, cluster + 1, fs);
+	    if (cluster != fs->data_clusters + 1)
+		get_fat(&subseqEntry, fs->fat, cluster + 1, fs);
+	    else
+		subseqEntry.value = 0;
 	    data[0] = new & 0xff;
-	    data[1] = (new >> 8) | (cluster == fs->clusters - 1 ? 0 :
-				    (0xff & subseqEntry.value) << 4);
+	    data[1] = (new >> 8) | ((0xff & subseqEntry.value) << 4);
 	}
 	size = 2;
 	break;
     case 16:
 	data = fs->fat + cluster * 2;
 	offs = fs->fat_start + cluster * 2;
-	*(unsigned short *)data = CT_LE_W(new);
+	*(unsigned short *)data = htole16(new);
 	size = 2;
 	break;
     case 32:
@@ -216,7 +320,7 @@ void set_fat(DOS_FS * fs, unsigned long cluster, unsigned long new)
 	    offs = fs->fat_start + cluster * 4;
 	    /* According to M$, the high 4 bits of a FAT32 entry are reserved and
 	     * are not part of the cluster number. So we never touch them. */
-	    *(unsigned long *)data = CT_LE_L((new & 0xfffffff) |
+	    *(uint32_t *)data = htole32((new & 0xfffffff) |
 					     (curEntry.reserved << 28));
 	    size = 4;
 	}
@@ -230,7 +334,7 @@ void set_fat(DOS_FS * fs, unsigned long cluster, unsigned long new)
     }
 }
 
-int bad_cluster(DOS_FS * fs, unsigned long cluster)
+int bad_cluster(DOS_FS * fs, uint32_t cluster)
 {
     FAT_ENTRY curEntry;
     get_fat(&curEntry, fs->fat, cluster, fs);
@@ -248,9 +352,9 @@ int bad_cluster(DOS_FS * fs, unsigned long cluster)
  * @return  -1              'cluster' is at the end of the chain
  * @return  Other values    Next cluster in this chain
  */
-unsigned long next_cluster(DOS_FS * fs, unsigned long cluster)
+uint32_t next_cluster(DOS_FS * fs, uint32_t cluster)
 {
-    unsigned long value;
+    uint32_t value;
     FAT_ENTRY curEntry;
 
     get_fat(&curEntry, fs->fat, cluster, fs);
@@ -261,10 +365,10 @@ unsigned long next_cluster(DOS_FS * fs, unsigned long cluster)
     return FAT_IS_EOF(fs, value) ? -1 : value;
 }
 
-loff_t cluster_start(DOS_FS * fs, unsigned long cluster)
+off_t cluster_start(DOS_FS * fs, uint32_t cluster)
 {
-    return fs->data_start + ((loff_t) cluster -
-			     2) * (unsigned long long)fs->cluster_size;
+    /* TODO: check overflow */
+    return fs->data_start + ((off_t)cluster - 2) * (unsigned long long)fs->cluster_size;
 }
 
 /**
@@ -276,7 +380,7 @@ loff_t cluster_start(DOS_FS * fs, unsigned long cluster)
  * @param[in]	    owner       Information on dentry that owns this cluster
  *                              (may be NULL)
  */
-void set_owner(DOS_FS * fs, unsigned long cluster, DOS_FILE * owner)
+void set_owner(DOS_FS * fs, uint32_t cluster, DOS_FILE * owner)
 {
     if (fs->cluster_owner == NULL)
 	die("Internal error: attempt to set owner in non-existent table");
@@ -287,7 +391,7 @@ void set_owner(DOS_FS * fs, unsigned long cluster, DOS_FILE * owner)
     fs->cluster_owner[cluster] = owner;
 }
 
-DOS_FILE *get_owner(DOS_FS * fs, unsigned long cluster)
+DOS_FILE *get_owner(DOS_FS * fs, uint32_t cluster)
 {
     if (fs->cluster_owner == NULL)
 	return NULL;
@@ -297,17 +401,17 @@ DOS_FILE *get_owner(DOS_FS * fs, unsigned long cluster)
 
 void fix_bad(DOS_FS * fs)
 {
-    unsigned long i;
+    uint32_t i;
 
     if (verbose)
 	printf("Checking for bad clusters.\n");
-    for (i = 2; i < fs->clusters + 2; i++) {
+    for (i = 2; i < fs->data_clusters + 2; i++) {
 	FAT_ENTRY curEntry;
 	get_fat(&curEntry, fs->fat, i, fs);
 
 	if (!get_owner(fs, i) && !FAT_IS_BAD(fs, curEntry.value))
 	    if (!fs_test(cluster_start(fs, i), fs->cluster_size)) {
-		printf("Cluster %lu is unreadable.\n", i);
+		printf("Cluster %lu is unreadable.\n", (unsigned long)i);
 		set_fat(fs, i, -2);
 	    }
     }
@@ -316,12 +420,12 @@ void fix_bad(DOS_FS * fs)
 void reclaim_free(DOS_FS * fs)
 {
     int reclaimed;
-    unsigned long i;
+    uint32_t i;
 
     if (verbose)
 	printf("Checking for unused clusters.\n");
     reclaimed = 0;
-    for (i = 2; i < fs->clusters + 2; i++) {
+    for (i = 2; i < fs->data_clusters + 2; i++) {
 	FAT_ENTRY curEntry;
 	get_fat(&curEntry, fs->fat, i, fs);
 
@@ -332,7 +436,7 @@ void reclaim_free(DOS_FS * fs)
 	}
     }
     if (reclaimed)
-	printf("Reclaimed %d unused cluster%s (%llu bytes).\n", reclaimed,
+	printf("Reclaimed %d unused cluster%s (%llu bytes).\n", (int)reclaimed,
 	       reclaimed == 1 ? "" : "s",
 	       (unsigned long long)reclaimed * fs->cluster_size);
 }
@@ -347,16 +451,16 @@ void reclaim_free(DOS_FS * fs)
  *				   clusters link to it.
  * @param[in]	    start_cluster  Where to start scanning for orphans
  */
-static void tag_free(DOS_FS * fs, DOS_FILE * owner, unsigned long *num_refs,
-		     unsigned long start_cluster)
+static void tag_free(DOS_FS * fs, DOS_FILE * owner, uint32_t *num_refs,
+		     uint32_t start_cluster)
 {
     int prev;
-    unsigned long i, walk;
+    uint32_t i, walk;
 
     if (start_cluster == 0)
 	start_cluster = 2;
 
-    for (i = start_cluster; i < fs->clusters + 2; i++) {
+    for (i = start_cluster; i < fs->data_clusters + 2; i++) {
 	FAT_ENTRY curEntry;
 	get_fat(&curEntry, fs->fat, i, fs);
 
@@ -400,16 +504,16 @@ void reclaim_file(DOS_FS * fs)
     DOS_FILE orphan;
     int reclaimed, files;
     int changed = 0;
-    unsigned long i, next, walk;
-    unsigned long *num_refs = NULL;	/* Only for orphaned clusters */
-    unsigned long total_num_clusters;
+    uint32_t i, next, walk;
+    uint32_t *num_refs = NULL;	/* Only for orphaned clusters */
+    uint32_t total_num_clusters;
 
     if (verbose)
 	printf("Reclaiming unconnected clusters.\n");
 
-    total_num_clusters = fs->clusters + 2UL;
-    num_refs = alloc(total_num_clusters * sizeof(unsigned long));
-    memset(num_refs, 0, (total_num_clusters * sizeof(unsigned long)));
+    total_num_clusters = fs->data_clusters + 2;
+    num_refs = alloc(total_num_clusters * sizeof(uint32_t));
+    memset(num_refs, 0, (total_num_clusters * sizeof(uint32_t)));
 
     /* Guarantee that all orphan chains (except cycles) end cleanly
      * with an end-of-chain mark.
@@ -420,7 +524,7 @@ void reclaim_file(DOS_FS * fs)
 	get_fat(&curEntry, fs->fat, i, fs);
 
 	next = curEntry.value;
-	if (!get_owner(fs, i) && next && next < fs->clusters + 2) {
+	if (!get_owner(fs, i) && next && next < fs->data_clusters + 2) {
 	    /* Cluster is linked, but not owned (orphan) */
 	    FAT_ENTRY nextEntry;
 	    get_fat(&nextEntry, fs->fat, next, fs);
@@ -454,7 +558,7 @@ void reclaim_file(DOS_FS * fs)
 		    die("Internal error: num_refs going below zero");
 		set_fat(fs, i, -1);
 		changed = curEntry.value;
-		printf("Broke cycle at cluster %lu in free chain.\n", i);
+		printf("Broke cycle at cluster %lu in free chain.\n", (unsigned long)i);
 
 		/* If we've created a new chain head,
 		 * tag_free() can claim it
@@ -472,15 +576,15 @@ void reclaim_file(DOS_FS * fs)
 	/* If this cluster is the head of an orphan chain... */
 	if (get_owner(fs, i) == &orphan && !num_refs[i]) {
 	    DIR_ENT de;
-	    loff_t offset;
+	    off_t offset;
 	    files++;
-	    offset = alloc_rootdir_entry(fs, &de, "FSCK%04d");
-	    de.start = CT_LE_W(i & 0xffff);
+	    offset = alloc_rootdir_entry(fs, &de, "FSCK%04dREC", 1);
+	    de.start = htole16(i & 0xffff);
 	    if (fs->fat_bits == 32)
-		de.starthi = CT_LE_W(i >> 16);
+		de.starthi = htole16(i >> 16);
 	    for (walk = i; walk > 0 && walk != -1;
 		 walk = next_cluster(fs, walk)) {
-		de.size = CT_LE_L(CF_LE_L(de.size) + fs->cluster_size);
+		de.size = htole32(le32toh(de.size) + fs->cluster_size);
 		reclaimed++;
 	    }
 	    fs_write(offset, sizeof(DIR_ENT), &de);
@@ -494,13 +598,13 @@ void reclaim_file(DOS_FS * fs)
     free(num_refs);
 }
 
-unsigned long update_free(DOS_FS * fs)
+uint32_t update_free(DOS_FS * fs)
 {
-    unsigned long i;
-    unsigned long free = 0;
+    uint32_t i;
+    uint32_t free = 0;
     int do_set = 0;
 
-    for (i = 2; i < fs->clusters + 2; i++) {
+    for (i = 2; i < fs->data_clusters + 2; i++) {
 	FAT_ENTRY curEntry;
 	get_fat(&curEntry, fs->fat, i, fs);
 
@@ -516,28 +620,26 @@ unsigned long update_free(DOS_FS * fs)
     if (fs->free_clusters != 0xFFFFFFFF) {
 	if (free != fs->free_clusters) {
 	    printf("Free cluster summary wrong (%ld vs. really %ld)\n",
-		   fs->free_clusters, free);
-	    if (interactive)
-		printf("1) Correct\n2) Don't correct\n");
-	    else
-		printf("  Auto-correcting.\n");
-	    if (!interactive || get_key("12", "?") == '1')
+		   (long)fs->free_clusters, (long)free);
+	    if (get_choice(1, "  Auto-correcting.",
+			   2,
+			   1, "Correct",
+			   2, "Don't correct") == 1)
 		do_set = 1;
 	}
     } else {
-	printf("Free cluster summary uninitialized (should be %ld)\n", free);
+	printf("Free cluster summary uninitialized (should be %ld)\n", (long)free);
 	if (rw) {
-	    if (interactive)
-		printf("1) Set it\n2) Leave it uninitialized\n");
-	    else
-		printf("  Auto-setting.\n");
-	    if (!interactive || get_key("12", "?") == '1')
+	    if (get_choice(1, "  Auto-setting.",
+			   2,
+			   1, "Set it",
+			   2, "Leave it uninitialized") == 1)
 		do_set = 1;
 	}
     }
 
     if (do_set) {
-	unsigned long le_free = CT_LE_L(free);
+	uint32_t le_free = htole32(free);
 	fs->free_clusters = free;
 	fs_write(fs->fsinfo_start + offsetof(struct info_sector, free_clusters),
 		 sizeof(le_free), &le_free);
